"""
===========================
Framework Utility Functions
===========================

Collection of utility functions shared by the ``vivarium`` framework.

"""
import functools
from bdb import BdbQuit
from importlib import import_module
from typing import Any, Callable

import numpy as np


def from_yearly(value, time_step):
    return value * (time_step.total_seconds() / (60 * 60 * 24 * 365.0))


def to_yearly(value, time_step):
    return value / (time_step.total_seconds() / (60 * 60 * 24 * 365.0))


def rate_to_probability(rate):
    # encountered underflow from rate > 30k
    # for rates greater than 250, exp(-rate) evaluates to 1e-109
    # beware machine-specific floating point issues
    rate[rate > 250] = 250.0
    return 1 - np.exp(-rate)


def probability_to_rate(probability):
    return -np.log(1 - probability)


def collapse_nested_dict(d, prefix=None):
    results = []
    for k, v in d.items():
        cur_prefix = prefix + "." + k if prefix else k
        if isinstance(v, dict):
            results.extend(collapse_nested_dict(v, prefix=cur_prefix))
        else:
            results.append((cur_prefix, v))
    return results


def import_by_path(path: str) -> Callable:
    """Import a class or function given it's absolute path.

    Parameters
    ----------
    path:
      Path to object to import
    """

    module_path, _, class_name = path.rpartition(".")
    return getattr(import_module(module_path), class_name)


def handle_exceptions(func: Callable, logger: Any, with_debugger: bool) -> Callable:
    """Drops a user into an interactive debugger if func raises an error."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (BdbQuit, KeyboardInterrupt):
            raise
        except Exception as e:
            logger.exception("Uncaught exception {}".format(e))
            if with_debugger:
                import pdb
                import traceback

                traceback.print_exc()
                pdb.post_mortem()
            else:
                raise

    return wrapped



"""
=============
Lookup Tables
=============

Simulations tend to require a large quantity of data to run.  :mod:`vivarium`
provides the :class:`LookupTable` abstraction to ensure that accurate data can
be retrieved when it's needed. It's a callable object that takes in a
population index and returns data specific to the individuals represented by
that index. See the :ref:`lookup concept note <lookup_concept>` for more.

"""
from datetime import datetime, timedelta
from numbers import Number
from typing import Callable, List, Tuple, Union

import pandas as pd

from vivarium.framework.population import PopulationView
from vivarium.interpolation import Interpolation

ScalarValue = Union[Number, timedelta, datetime]


class InterpolatedTable:
    """A callable that interpolates data according to a given strategy.

    Attributes
    ----------
    data
        The data from which to build the interpolation.
    population_view
        View of the population to be used when the table is called with an
        index.
    key_columns
        Column names to be used as categorical parameters in Interpolation
        to select between interpolation functions.
    parameter_columns
        Column names to be used as continuous parameters in Interpolation.
    value_columns
        Names of value columns to be interpolated over. All non parameter- and
        key- columns in data.
    interpolation_order
        Order of interpolation. Used to decide interpolation strategy.
    clock
        Callable for current time in simulation.
    extrapolate
        Whether or not to extrapolate beyond edges of given bins.

    Notes
    -----
    These should not be created directly. Use the `lookup` interface on the
    :class:`builder <vivarium.framework.engine.Builder>` during setup.

    """

    def __init__(
        self,
        data: pd.DataFrame,
        population_view: PopulationView,
        key_columns: Union[List[str], Tuple[str]],
        parameter_columns: Union[List[str], Tuple],
        value_columns: Union[List[str], Tuple[str]],
        interpolation_order: int,
        clock: Callable,
        extrapolate: bool,
        validate: bool,
    ):

        self.data = data
        self.population_view = population_view
        self.key_columns = key_columns
        param_cols_with_edges = []
        for p in parameter_columns:
            param_cols_with_edges += [(p, f"{p}_start", f"{p}_end")]
        self.parameter_columns = param_cols_with_edges
        self.interpolation_order = interpolation_order
        self.value_columns = value_columns
        self.clock = clock
        self.extrapolate = extrapolate
        self.validate = validate
        self.interpolation = Interpolation(
            data,
            self.key_columns,
            self.parameter_columns,
            order=self.interpolation_order,
            extrapolate=self.extrapolate,
            validate=self.validate,
        )

    def __call__(self, index: pd.Index) -> pd.DataFrame:
        """Get the interpolated values for the rows in ``index``.

        Parameters
        ----------
        index
            Index of the population to interpolate for.

        Returns
        -------
            A table with the interpolated values for the population requested.

        """
        pop = self.population_view.get(index)
        del pop["tracked"]
        if "year" in [col for p in self.parameter_columns for col in p]:
            current_time = self.clock()
            fractional_year = current_time.year
            fractional_year += current_time.timetuple().tm_yday / 365.25
            pop["year"] = fractional_year

        return self.interpolation(pop)

    def __repr__(self):
        return "InterpolatedTable()"


class ScalarTable:
    """A callable that broadcasts a scalar or list of scalars over an index.

    Attributes
    ----------
    values
        The scalar value(s) from which to build table columns.
    value_columns
        List of string names to be used to name the columns of the table built
        from values.

    Notes
    -----
    These should not be created directly. Use the `lookup` interface on the
    builder during setup.

    """

    def __init__(
        self,
        values: Union[List[ScalarValue], Tuple[ScalarValue]],
        value_columns: Union[List[str], Tuple[str]],
    ):

        self.values = values
        self.value_columns = value_columns

    def __call__(self, index: pd.Index) -> pd.DataFrame:
        """Broadcast this tables values over the provided index.

        Parameters
        ----------
        index
            Index of the population to construct table for.

        Returns
        -------
            A table with a column for each of the scalar values for the
            population requested.

        """
        if not isinstance(self.values, (list, tuple)):
            values = pd.Series(
                self.values,
                index=index,
                name=self.value_columns[0] if self.value_columns else None,
            )
        else:
            values = dict(
                zip(self.value_columns, [pd.Series(v, index=index) for v in self.values])
            )
        return pd.DataFrame(values)

    def __repr__(self):
        return "ScalarTable(value(s)={})".format(self.values)


class LookupTable:
    """Wrapper for different strategies for looking up values for an index.

    In :mod:`vivarium` simulations, the index is synonymous with the simulated
    population.  The lookup system allows the user to provide different kinds
    of data and strategies for using that data.  When the simulation is
    running, then, components can lookup parameter values based solely on
    the population index.

    Notes
    -----
    These should not be created directly. Use the `lookup` method on the builder
    during setup.

    """

    def __init__(
        self,
        table_number: int,
        data: Union[ScalarValue, pd.DataFrame, List[ScalarValue], Tuple[ScalarValue]],
        population_view: Callable,
        key_columns: Union[List[str], Tuple[str]],
        parameter_columns: Union[List[str], Tuple],
        value_columns: Union[List[str], Tuple[str]],
        interpolation_order: int,
        clock: Callable,
        extrapolate: bool,
        validate: bool,
    ):
        self.table_number = table_number
        key_columns = [] if key_columns is None else key_columns

        if validate:
            validate_parameters(data, key_columns, parameter_columns, value_columns)

        # Note datetime catches pandas timestamps
        if isinstance(data, (Number, datetime, timedelta, list, tuple)):
            self._table = ScalarTable(data, value_columns)
        else:
            view_columns = sorted((set(key_columns) | set(parameter_columns)) - {"year"}) + [
                "tracked"
            ]
            self._table = InterpolatedTable(
                data,
                population_view(view_columns),
                key_columns,
                parameter_columns,
                value_columns,
                interpolation_order,
                clock,
                extrapolate,
                validate,
            )

    @property
    def name(self) -> str:
        """Tables are generically named after the order they were created."""
        return f"lookup_table_{self.table_number}"

    def __call__(self, index: pd.Index) -> Union[pd.Series, pd.DataFrame]:
        """Get the interpolated or scalar table values for the given index.

        Parameters
        ----------
        index
            Index for population view.

        Returns
        -------
            pandas.Series if interpolated or scalar values for index are one
            column, pandas.DataFrame if multiple columns

        """
        return self._call(index)

    def _call(self, index: pd.Index) -> Union[pd.Series, pd.DataFrame]:
        """Private method to allow LookupManager to add constraints."""
        table_view = self._table(index)
        if len(table_view.columns) == 1:
            return table_view[table_view.columns[0]]
        return table_view

    def __repr__(self):
        return "LookupTable()"


def validate_parameters(
    data: pd.DataFrame,
    key_columns: Union[List[str], Tuple[str]],
    parameter_columns: Union[List[str], Tuple],
    value_columns: Union[List[str], Tuple[str]],
):
    """Makes sure the data format agrees with the provided column layout."""
    if (
        data is None
        or (isinstance(data, pd.DataFrame) and data.empty)
        or (isinstance(data, (list, tuple)) and not data)
    ):
        raise ValueError("Must supply some data")

    if not isinstance(data, (Number, datetime, timedelta, list, tuple, pd.DataFrame)):
        raise TypeError(
            f"The only allowable types for data are number, datetime, timedelta, "
            f"list, tuple, or pandas.DataFrame. You passed {type(data)}."
        )

    if isinstance(data, (list, tuple)):
        if not value_columns:
            raise ValueError(
                f"To invoke scalar view with multiple values, you must supply value_columns"
            )
        if len(value_columns) != len(data):
            raise ValueError(
                f"The number of value columns must match the number of values."
                f"You supplied values: {data} and value_columns: {value_columns}"
            )

    if isinstance(data, pd.DataFrame):
        all_parameter_columns = []
        for p in parameter_columns:
            all_parameter_columns += [p, f"{p}_start", f"{p}_end"]
        if set(key_columns).intersection(set(all_parameter_columns)):
            raise ValueError(
                f"There should be no overlap between key columns: {key_columns} "
                f"and parameter columns: {parameter_columns}."
            )

        if value_columns:
            data_value_columns = data.columns.difference(
                set(key_columns) | set(all_parameter_columns)
            )
            if set(value_columns) != set(data_value_columns):
                raise ValueError(
                    f"The value columns you supplied: {value_columns} do not match "
                    f"the non-parameter columns in the passed data: {data_value_columns}"
                )


class LookupTableManager:
    """Manages complex data in the simulation.

    Notes
    -----
    Client code should never access this class directly. Use ``lookup`` on the
    builder during setup to get references to LookupTable objects.

    """

    configuration_defaults = {
        "interpolation": {"order": 0, "validate": True, "extrapolate": True}
    }

    @property
    def name(self):
        return "lookup_table_manager"

    def setup(self, builder):
        self.tables = {}
        self._pop_view_builder = builder.population.get_view
        self.clock = builder.time.clock()
        self._interpolation_order = builder.configuration.interpolation.order
        self._extrapolate = builder.configuration.interpolation.extrapolate
        self._validate = builder.configuration.interpolation.validate
        self._add_constraint = builder.lifecycle.add_constraint

        builder.lifecycle.add_constraint(self.build_table, allow_during=["setup"])

    def build_table(
        self,
        data: Union[ScalarValue, pd.DataFrame, List[ScalarValue], Tuple[ScalarValue]],
        key_columns: Union[List[str], Tuple[str]],
        parameter_columns: Union[List[str], Tuple[str]],
        value_columns: Union[List[str], Tuple[str]],
    ) -> LookupTable:
        """Construct a lookup table from input data."""
        table = self._build_table(data, key_columns, parameter_columns, value_columns)
        self._add_constraint(
            table._call, restrict_during=["initialization", "setup", "post_setup"]
        )
        return table

    def _build_table(self, data, key_columns, parameter_columns, value_columns):
        # We don't want to require explicit names for tables, but giving them
        # generic names is useful for introspection.
        table_number = len(self.tables)
        table = LookupTable(
            table_number,
            data,
            self._pop_view_builder,
            key_columns,
            parameter_columns,
            value_columns,
            self._interpolation_order,
            self.clock,
            self._extrapolate,
            self._validate,
        )
        self.tables[table_number] = table
        return table

    def __repr__(self):
        return "LookupTableManager()"


class LookupTableInterface:
    """The lookup table management system.

    Simulations tend to require a large quantity of data to run. ``vivarium``
    provides the :class:`Lookup Table <vivarium.framework.lookup.LookupTable>`
    abstraction to ensure that accurate data can be retrieved when it's needed.

    For more information, see :ref:`here <lookup_concept>`.

    """

    def __init__(self, manager: LookupTableManager):
        self._manager = manager

    def build_table(
        self,
        data: Union[ScalarValue, pd.DataFrame, List[ScalarValue], Tuple[ScalarValue]],
        key_columns: Union[List[str], Tuple[str]] = None,
        parameter_columns: Union[List[str], Tuple[str]] = None,
        value_columns: Union[List[str], Tuple[str]] = None,
    ) -> LookupTable:
        """Construct a LookupTable from input data.

        If data is a :class:`pandas.DataFrame`, an interpolation function of
        the order specified in the simulation
        :term:`configuration <Configuration>` will be calculated for each
        permutation of the set of key_columns. The columns in parameter_columns
        will be used as parameters for the interpolation functions which will
        estimate all remaining columns in the table.

        If data is a number, time, list, or tuple, a scalar table will be
        constructed with the values in data as the values in each column of
        the table, named according to value_columns.


        Parameters
        ----------
        data
            The source data which will be used to build the resulting
            :class:`LookupTable`.
        key_columns
            Columns used to select between interpolation functions. These
            should be the non-continuous variables in the data. For example
            'sex' in data about a population.
        parameter_columns
            The columns which contain the parameters to the interpolation
            functions. These should be the continuous variables. For example
            'age' in data about a population.
        value_columns
            The data columns that will be in the resulting LookupTable. Columns
            to be interpolated over if interpolation or the names of the columns
            in the scalar table.

        Returns
        -------
            LookupTable
        """
        return self._manager.build_table(data, key_columns, parameter_columns, value_columns)



import math
from datetime import timedelta

import pandas as pd
import pytest

from vivarium.framework.results.context import ResultsContext

from .mocks import (
    BASE_POPULATION,
    CATEGORIES,
    FAMILIARS,
    NAME,
    SOURCES,
    sorting_hat_serial,
    sorting_hat_vector,
    verify_stratification_added,
)


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized",
    [
        (NAME, SOURCES, CATEGORIES, sorting_hat_vector, True),
        (NAME, SOURCES, CATEGORIES, sorting_hat_serial, False),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_add_stratification(name, sources, categories, mapper, is_vectorized):
    ctx = ResultsContext()
    assert not verify_stratification_added(
        ctx._stratifications, name, sources, categories, mapper, is_vectorized
    )
    ctx.add_stratification(name, sources, categories, mapper, is_vectorized)
    assert verify_stratification_added(
        ctx._stratifications, name, sources, categories, mapper, is_vectorized
    )


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized, expected_exception",
    [
        (  # sources not in population columns
            NAME,
            ["middle_initial"],
            CATEGORIES,
            sorting_hat_vector,
            True,
            TypeError,
        ),
        (  # is_vectorized=True with non-vectorized mapper
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_serial,
            True,
            Exception,
        ),
        (  # is_vectorized=False with vectorized mapper
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_vector,
            False,
            Exception,
        ),
    ],
)
def test_add_stratification_raises(
    name, sources, categories, mapper, is_vectorized, expected_exception
):
    ctx = ResultsContext()
    with pytest.raises(expected_exception):
        raise ctx.add_stratification(name, sources, categories, mapper, is_vectorized)


def _aggregate_state_person_time(x: pd.DataFrame) -> float:
    """Helper aggregator function for observation testing"""
    return len(x) * (28 / 365.35)


@pytest.mark.parametrize(
    "name, pop_filter, aggregator, additional_stratifications, excluded_stratifications, when",
    [
        (
            "living_person_time",
            'alive == "alive" and undead == False',
            _aggregate_state_person_time,
            [],
            [],
            "collect_metrics",
        ),
        (
            "undead_person_time",
            "undead == True",
            _aggregate_state_person_time,
            [],
            [],
            "time_step__prepare",
        ),
    ],
    ids=["valid_on_collect_metrics", "valid_on_time_step__prepare"],
)
def test_add_observation(
    name, pop_filter, aggregator, additional_stratifications, excluded_stratifications, when
):
    ctx = ResultsContext()
    ctx._default_stratifications = ["age", "sex"]
    assert len(ctx._observations) == 0
    ctx.add_observation(
        name,
        pop_filter,
        [],
        aggregator,
        additional_stratifications,
        excluded_stratifications,
        when,
    )
    assert len(ctx._observations) == 1


@pytest.mark.parametrize(
    "name, pop_filter, aggregator, additional_stratifications, excluded_stratifications, when",
    [
        (
            "living_person_time",
            'alive == "alive" and undead == False',
            _aggregate_state_person_time,
            [],
            [],
            "collect_metrics",
        ),
    ],
    ids=["valid_on_collect_metrics"],
)
def test_double_add_observation(
    name, pop_filter, aggregator, additional_stratifications, excluded_stratifications, when
):
    """Tests a double add of the same stratification, this should result in one additional observation being added to
    the context."""
    ctx = ResultsContext()
    ctx._default_stratifications = ["age", "sex"]
    assert len(ctx._observations) == 0
    ctx.add_observation(
        name,
        pop_filter,
        [],
        aggregator,
        additional_stratifications,
        excluded_stratifications,
        when,
    )
    ctx.add_observation(
        name,
        pop_filter,
        [],
        aggregator,
        additional_stratifications,
        excluded_stratifications,
        when,
    )
    assert len(ctx._observations) == 1


@pytest.mark.parametrize(
    "default, additional, excluded, match",
    [
        (["age", "sex"], ["age"], [], "age"),
        (["age", "sex"], [], ["eye_color"], "eye_color"),
        (["age", "sex"], ["age"], ["eye_color"], "age|eye_color"),
    ],
    ids=[
        "additional_no_operation",
        "exclude_no_operation",
        "additional_and_exclude_no_operation",
    ],
)
def test_add_observation_nop_stratifications(default, additional, excluded, match):
    ctx = ResultsContext()
    ctx._default_stratifications = default
    with pytest.warns(UserWarning, match=match):
        ctx.add_observation(
            "name",
            'alive == "alive"',
            [],
            _aggregate_state_person_time,
            additional,
            excluded,
            "collect_metrics",
        )


@pytest.mark.parametrize(
    "default_stratifications, additional_stratifications, excluded_stratifications, expected_stratifications",
    [
        ([], [], [], ()),
        (["age", "sex"], ["handedness"], ["age"], ("sex", "handedness")),
        (["age", "sex"], [], ["age", "sex"], ()),
        (["age"], [], ["bogus_exclude_column"], ("age",)),
    ],
    ids=[
        "empty_add_empty_exclude",
        "one_add_one_exclude",
        "all_defaults_excluded",
        "bogus_exclude",
    ],
)
def test__get_stratifications(
    default_stratifications,
    additional_stratifications,
    excluded_stratifications,
    expected_stratifications,
):
    ctx = ResultsContext()
    # default_stratifications would normally be set via ResultsInterface.set_default_stratifications()
    ctx._default_stratifications = default_stratifications
    stratifications = ctx._get_stratifications(
        additional_stratifications, excluded_stratifications
    )
    assert sorted(stratifications) == sorted(expected_stratifications)


@pytest.mark.parametrize(
    "name, pop_filter, aggregator_sources, aggregator, stratifications, expected_result",
    [
        ("wizard_count", "tracked==True", None, len, ["house", "familiar"], 4),
        (
            "power_level_total",
            "tracked==True",
            ["power_level"],
            sum,
            ["house", "familiar"],
            260,
        ),
        (
            "wizard_time",
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["house", "familiar"],
            0.306555,
        ),
        ("wizard_count", "tracked==True", None, len, ["house"], 20),
        ("power_level_total", "tracked==True", ["power_level"], sum, ["house"], 1300),
        (
            "wizard_time",
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["house"],
            1.53277,
        ),
    ],
    ids=[
        "len_aggregator_two_stratifications",
        "sum_aggregator_two_stratifications",
        "custom_aggregator_two_stratifications",
        "len_aggregator_one_stratification",
        "sum_aggregator_one_stratification",
        "custom_aggregator_one_stratification",
    ],
)
def test_gather_results(
    name, pop_filter, aggregator_sources, aggregator, stratifications, expected_result
):
    """Test cases where every stratification is in gather_results. Checks for existence and correctness
    of results"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )
    event_name = "collect_metrics"

    # Set up stratifications
    if "house" in stratifications:
        ctx.add_stratification("house", ["house"], CATEGORIES, None, True)
    if "familiar" in stratifications:
        ctx.add_stratification("familiar", ["familiar"], FAMILIARS, None, True)
    ctx.add_observation(
        name,
        pop_filter,
        aggregator_sources,
        aggregator,
        stratifications,
        [],
        event_name,
    )

    i = 0
    for r in ctx.gather_results(population, event_name):
        assert all(
            math.isclose(result, expected_result, rel_tol=0.0001) for result in r.values()
        )
        i += 1
    assert i == 1


@pytest.mark.parametrize(
    "name, pop_filter, aggregator_sources, aggregator, stratifications",
    [
        ("wizard_count", "tracked==True", None, len, ["house", "familiar"]),
        ("power_level_total", "tracked==True", ["power_level"], sum, ["house", "familiar"]),
        (
            "wizard_time",
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["house", "familiar"],
        ),
        ("wizard_count", "tracked==True", None, len, ["familiar"]),
        ("power_level_total", "tracked==True", ["power_level"], sum, ["familiar"]),
        (
            "wizard_time",
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["familiar"],
        ),
    ],
    ids=[
        "len_aggregator_two_stratifications",
        "sum_aggregator_two_stratifications",
        "custom_aggregator_two_stratifications",
        "len_aggregator_one_stratification",
        "sum_aggregator_one_stratification",
        "custom_aggregator_one_stratification",
    ],
)
def test_gather_results_partial_stratifications_in_results(
    name, pop_filter, aggregator_sources, aggregator, stratifications
):
    """Test cases where not all stratifications are observed for gather_results. This looks for existence of
    unobserved stratifications and ensures their values are 0"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )
    # Remove an entire category from a stratification
    population = population[population["familiar"] != "unladen_swallow"].reset_index()

    event_name = "collect_metrics"

    # Set up stratifications
    if "house" in stratifications:
        ctx.add_stratification("house", ["house"], CATEGORIES, None, True)
    if "familiar" in stratifications:
        ctx.add_stratification("familiar", ["familiar"], FAMILIARS, None, True)
    ctx.add_observation(
        name,
        pop_filter,
        aggregator_sources,
        aggregator,
        stratifications,
        [],
        event_name,
    )

    for r in ctx.gather_results(population, event_name):
        unladen_results = {k: v for (k, v) in r.items() if "unladen_swallow" in k}
        assert len(unladen_results.items()) > 0
        assert all(v == 0 for v in unladen_results.values())


def test__format_results():
    """Test that format results produces the expected number of keys and a specific expected key"""
    ctx = ResultsContext()
    aggregates = BASE_POPULATION.groupby(["house", "familiar"]).apply(len)
    measure = "wizard_count"
    rv = ctx._format_results(measure, aggregates)

    # Check that the number of expected data column names are there
    expected_keys_len = len(CATEGORIES) * len(FAMILIARS)
    assert len(rv.keys()) == expected_keys_len

    # Check that an example data column name is there
    expected_key = "MEASURE_wizard_count_HOUSE_slytherin_FAMILIAR_cat"
    assert expected_key in rv.keys()


def test__bad_aggregator_return():
    """Test that an exception is raised, as expected, when an aggregator
    produces something other than a pd.DataFrame with a single column or a pd.Series"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    event_name = "collect_metrics"

    # Set up stratifications
    ctx.add_stratification("house", ["house"], CATEGORIES, None, True)
    ctx.add_stratification("familiar", ["familiar"], FAMILIARS, None, True)
    ctx.add_observation(
        "this_shouldnt_work",
        "",
        ["tracked", "power_level"],
        sum,
        ["house", "familiar"],
        [],
        event_name,
    )

    with pytest.raises(TypeError):
        for r in ctx.gather_results(population, event_name):
            print(r)


def test__bad_aggregator_stratification():
    """Test if an exception gets raised when a stratification that doesn't
    exist is attempted to be used, as expected."""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    event_name = "collect_metrics"

    # Set up stratifications
    ctx.add_stratification("house", ["house"], CATEGORIES, None, True)
    ctx.add_stratification("familiar", ["familiar"], FAMILIARS, None, True)
    ctx.add_observation(
        "this_shouldnt_work",
        "",
        [],
        sum,
        ["house", "height"],  # `height` is not a stratification
        [],
        event_name,
    )

    with pytest.raises(KeyError, match="height"):
        for r in ctx.gather_results(population, event_name):
            print(r)



from .artifact import Artifact, ArtifactException
from .hdf import EntityKey
from .manager import (
    ArtifactInterface,
    ArtifactManager,
    filter_data,
    parse_artifact_path_config,
    validate_filter_term,
)



class MockComponentA:
    def __init__(self, *args, name="mock_component_a"):
        self.name = name
        self.args = args
        self.builder_used_for_setup = None

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name


class MockComponentB:
    def __init__(self, *args, name="mock_component_b"):
        self._name = name
        self.args = args
        self.builder_used_for_setup = None
        self._sub_components = []
        if len(self.args) > 1:
            for arg in self.args:
                self._sub_components.append(MockComponentB(arg, name=arg))

    @property
    def name(self):
        return self._name

    @property
    def sub_components(self):
        return self._sub_components

    def setup(self, builder):
        self.builder_used_for_setup = builder
        builder.value.register_value_modifier("metrics", self.metrics)

    def metrics(self, _, metrics):
        if "test" in metrics:
            metrics["test"] += 1
        else:
            metrics["test"] = 1
        return metrics

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name


class MockGenericComponent:

    configuration_defaults = {
        "component": {
            "key1": "val",
            "key2": {
                "subkey1": "val",
                "subkey2": "val",
            },
            "key3": ["val", "val", "val"],
        }
    }

    def __init__(self, name):
        self.name = name
        self.configuration_defaults = {
            self.name: MockGenericComponent.configuration_defaults["component"]
        }
        self.builder_used_for_setup = None

    def setup(self, builder):
        self.builder_used_for_setup = builder
        self.config = builder.configuration[self.name]

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name


class NamelessComponent:
    def __init__(self, *args):
        self.args = args


class Listener(MockComponentB):
    def __init__(self, *args, name="test_listener"):
        super().__init__(*args, name=name)
        self.post_setup_called = False
        self.time_step__prepare_called = False
        self.time_step_called = False
        self.time_step__cleanup_called = False
        self.collect_metrics_called = False
        self.simulation_end_called = False

    def setup(self, builder):
        super().setup(builder)
        builder.event.register_listener("post_setup", self.on_post_setup)
        builder.event.register_listener("time_step__prepare", self.on_time_step__prepare)
        builder.event.register_listener("time_step", self.on_time_step)
        builder.event.register_listener("time_step__cleanup", self.on_time_step__cleanup)
        builder.event.register_listener("collect_metrics", self.on_collect_metrics)
        builder.event.register_listener("simulation_end", self.on_simulation_end)

    def on_post_setup(self, _):
        self.post_setup_called = True

    def on_time_step__prepare(self, _):
        self.time_step__prepare_called = True

    def on_time_step(self, _):
        self.time_step_called = True

    def on_time_step__cleanup(self, _):
        self.time_step__cleanup_called = True

    def on_collect_metrics(self, _):
        self.collect_metrics_called = True

    def on_simulation_end(self, _):
        self.simulation_end_called = True



from dataclasses import dataclass
from typing import Callable, List, Union

import pandas as pd
from pandas.api.types import CategoricalDtype


@dataclass
class Stratification:
    """Class for stratifying observed quantities by specified characteristics

    Each Stratification represents a set of mutually exclusive and collectively
    exhaustive categories into which simulants can be assigned.

    The `Stratification` class has five fields: `name`, `source`, `mapper`,
    `categories`, and `is_vectorized`. The `name` is the name of the column
    created by the mapper. The `source` is a list of columns in the extended
    state table that are the inputs to the mapper function.  Simulants will
    later be grouped by this column (or these columns) during stratification.
    `categories` is a set of values that the mapper is allowed to output. The
    `mapper` is the method that transforms the source to the name column.
    The method produces an output column by calling the mapper on the source
    columns. If the mapper is `None`, the default identity mapper is used. If
    the mapper is not vectorized this is performed by using `pd.apply`.
    Finally, `is_vectorized` is a boolean parameter that signifies whether
    mapper function is applied to a single simulant (`False`) or to the whole
    population (`True`).

    `Stratification` also has a `__call__()` method. The method produces an
    output column by calling the mapper on the source columns.
    """

    name: str
    sources: List[str]
    categories: List[str]
    mapper: Callable[[Union[pd.Series, pd.DataFrame]], Union[str, pd.Series]] = None
    is_vectorized: bool = False

    def __post_init__(self):
        if self.mapper is None:
            if len(self.sources) != 1:
                raise ValueError(
                    f"No mapper provided for stratification {self.name} with"
                    f"{len(self.sources)} stratification sources."
                )
            self.mapper = self._default_mapper
            self.is_vectorized = True
        if not len(self.categories):
            raise ValueError("The categories argument must be non-empty.")
        if not len(self.sources):
            raise ValueError("The sources argument must be non-empty.")

    def __call__(self, population: pd.DataFrame) -> pd.DataFrame:
        if self.is_vectorized:
            raw_mapped_column = self.mapper(population[self.sources])
        else:
            raw_mapped_column = population[self.sources].apply(self.mapper, axis=1)
        mapped_column = raw_mapped_column.astype(
            CategoricalDtype(categories=self.categories, ordered=True)
        )
        if mapped_column.isna().any():
            invalid_categories = set(raw_mapped_column.unique()) - set(self.categories)
            raise ValueError(f"Invalid values '{invalid_categories}' found in {self.name}.")

        population[self.name] = mapped_column
        return population

    @staticmethod
    def _default_mapper(pop: pd.DataFrame) -> str:
        return pop.squeeze(axis=1)



"""
===========================
Vivarium Command Line Tools
===========================

``vivarium`` provides the tool :command:`simulate` for running simulations
from the command line.  It provides three subcommands:

.. list-table:: ``simulate`` sub-commands
    :header-rows: 1
    :widths: 30, 40

    *   - Name
        - Description
    *   - | **run**
        - | Runs a single simulation from a model specification file.
    *   - | **test**
        - | Runs an example simulation that comes packaged with ``vivarium``.
           | Useful as an installation test.
    *   - | **profile**
        - | Produces a profile of a simulation using the python
          | :mod:`cProfile` module

For more information, see the :ref:`tutorial <cli_tutorial>` on running
simulations from the command line.

.. click:: vivarium.interface.cli:simulate
   :prog: simulate
   :show-nested:

"""
import cProfile
import os
import pstats
import shutil
from pathlib import Path
from time import time

import click
import pandas as pd
from loguru import logger

from vivarium.examples import disease_model
from vivarium.framework.engine import run_simulation
from vivarium.framework.utilities import handle_exceptions

from .utilities import (
    configure_logging_to_file,
    configure_logging_to_terminal,
    get_output_root,
)


@click.group()
def simulate():
    """A command line utility for running a single simulation.

    You may initiate a new run with the ``run`` sub-command, initiate a test
    run of a provided model specification with the ``test`` subcommand, or
    profile a simulation run with the ``profile`` subcommand.
    """
    pass


@simulate.command()
@click.argument(
    "model_specification", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option("--location", "-l", help="Location to run the simulation in.")
@click.option(
    "--artifact_path",
    "-i",
    type=click.Path(resolve_path=True),
    help="The path to the artifact data file.",
)
@click.option(
    "--results_directory",
    "-o",
    type=click.Path(resolve_path=True),
    default=Path("~/vivarium_results/").expanduser(),
    help="The directory to write results to. A folder will be created "
    "in this directory with the same name as the configuration file.",
)
@click.option("--verbose", "-v", is_flag=True, help="Report each time step.")
@click.option(
    "--pdb",
    "with_debugger",
    is_flag=True,
    help="Drop into python debugger if an error occurs.",
)
def run(
    model_specification: Path,
    location: str,
    artifact_path: Path,
    results_directory: Path,
    verbose: int,
    with_debugger: bool,
):
    """Run a simulation from the command line.

    The simulation itself is defined by the given MODEL_SPECIFICATION yaml file.

    Within the results directory, which defaults to ~/vivarium_results if none
    is provided, a subdirectory will be created with the same name as the
    MODEL_SPECIFICATION if one does not exist. Results will be written to a
    further subdirectory named after the start time of the simulation run."""
    configure_logging_to_terminal(verbose)

    start = time()

    results_root = get_output_root(results_directory, model_specification, artifact_path)
    # Update permissions mask (assign to variable to avoid printing previous value)
    _ = os.umask(0o002)
    results_root.mkdir(parents=True, exist_ok=False)

    configure_logging_to_file(results_root)
    shutil.copy(model_specification, results_root / "model_specification.yaml")

    output_data = {"results_directory": str(results_root)}
    input_data = {}
    if location:
        input_data["location"] = location
    if artifact_path:
        input_data["artifact_path"] = artifact_path
    override_configuration = {"output_data": output_data, "input_data": input_data}

    main = handle_exceptions(run_simulation, logger, with_debugger)
    finished_sim = main(model_specification, configuration=override_configuration)

    metrics = pd.DataFrame(finished_sim.report(), index=[0])
    metrics["simulation_run_time"] = time() - start
    metrics["random_seed"] = finished_sim.configuration.randomness.random_seed
    metrics["input_draw"] = finished_sim.configuration.input_data.input_draw_number
    metrics.to_hdf(results_root / "output.hdf", key="data")


@simulate.command()
def test():
    """Run a test simulation using the ``disease_model.yaml`` model specification
    provided in the examples directory.
    """
    configure_logging_to_terminal(verbose=True)
    model_specification = disease_model.get_model_specification_path()

    main = handle_exceptions(run_simulation, logger, with_debugger=False)

    main(model_specification)
    click.echo()
    click.secho("Installation test successful!", fg="green")


@simulate.command()
@click.argument(
    "model_specification", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    "--results_directory",
    "-o",
    type=click.Path(resolve_path=True),
    default=Path("~/vivarium_results/").expanduser(),
    help="The directory to write results to. A folder will be created "
    "in this directory with the same name as the configuration file.",
)
@click.option(
    "--process/--no-process",
    default=False,
    help=(
        "Automatically process the profile to human readable format  with pstats, "
        "sorted by cumulative runtime, and dump to a file"
    ),
)
def profile(model_specification, results_directory, process):
    """Run a simulation based on the provided MODEL_SPECIFICATION and profile
    the run.
    """
    model_specification = Path(model_specification)
    results_directory = Path(results_directory)

    out_stats_file = results_directory / f"{model_specification.name}".replace(
        "yaml", "stats"
    )
    command = f'run_simulation("{model_specification}")'
    cProfile.runctx(command, globals=globals(), locals=locals(), filename=out_stats_file)

    if process:
        out_txt_file = results_directory / (out_stats_file.name + ".txt")
        with out_txt_file.open("w") as f:
            p = pstats.Stats(str(out_stats_file), stream=f)
            p.sort_stats("cumulative")
            p.print_stats()



from types import MethodType

import pytest

from vivarium.framework.results.manager import ResultsManager

from .mocks import (
    BIN_BINNED_COLUMN,
    BIN_LABELS,
    BIN_SILLY_BINS,
    BIN_SOURCE,
    CATEGORIES,
    NAME,
    SOURCES,
    sorting_hat_serial,
    sorting_hat_vector,
    verify_stratification_added,
)


# Mock for get_value call for Pipelines, returns a str instead of a Pipeline
def mock_get_value(self, name: str):
    return name


#######################################
# Tests for `register_stratification` #
#######################################


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized",
    [
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_vector,
            True,
        ),
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_serial,
            False,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_no_pipelines(
    name, sources, categories, mapper, is_vectorized, mocker
):
    mgr = ResultsManager()
    builder = mocker.Mock()
    mgr.setup(builder)
    mgr.register_stratification(name, categories, mapper, is_vectorized, sources, [])
    for item in sources:
        assert item in mgr._required_columns
    assert verify_stratification_added(
        mgr._results_context._stratifications,
        name,
        sources,
        categories,
        mapper,
        is_vectorized,
    )


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized",
    [
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_vector,
            True,
        ),
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_serial,
            False,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_with_pipelines(
    name, sources, categories, mapper, is_vectorized, mocker
):
    mgr = ResultsManager()
    builder = mocker.Mock()
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    mgr.register_stratification(name, categories, mapper, is_vectorized, [], sources)
    for item in sources:
        assert item in mgr._required_values
    assert verify_stratification_added(
        mgr._results_context._stratifications,
        name,
        sources,
        categories,
        mapper,
        is_vectorized,
    )


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized",
    [
        (  # expected Stratification for vectorized
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_vector,
            True,
        ),
        (  # expected Stratification for non-vectorized
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_serial,
            False,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_with_column_and_pipelines(
    name, sources, categories, mapper, is_vectorized, mocker
):
    mgr = ResultsManager()
    builder = mocker.Mock()
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    mocked_column_name = "silly_column"
    mgr.register_stratification(
        name, categories, mapper, is_vectorized, [mocked_column_name], sources
    )
    assert mocked_column_name in mgr._required_columns
    for item in sources:
        assert item in mgr._required_values
    all_sources = sources.copy()
    all_sources.append(mocked_column_name)
    assert verify_stratification_added(
        mgr._results_context._stratifications,
        name,
        all_sources,
        categories,
        mapper,
        is_vectorized,
    )


def test_duplicate_name_register_stratification(mocker):
    mgr = ResultsManager()
    builder = mocker.Mock()
    mgr.setup(builder)
    mgr.register_stratification(NAME, CATEGORIES, sorting_hat_serial, False, SOURCES, [])
    with pytest.raises(ValueError, match=f"Name `{NAME}` is already used"):
        mgr.register_stratification(NAME, CATEGORIES, sorting_hat_vector, True, SOURCES, [])


##############################################
# Tests for `register_binned_stratification` #
##############################################


def test_register_binned_stratification(mocker):
    mgr = ResultsManager()
    mock_register_stratification = mocker.patch(
        "vivarium.framework.results.manager.ResultsManager.register_stratification"
    )
    mgr.register_binned_stratification(
        BIN_SOURCE, "column", BIN_BINNED_COLUMN, BIN_SILLY_BINS, BIN_LABELS
    )
    mock_register_stratification.assert_called_once()


@pytest.mark.parametrize(
    "bins, labels",
    [(BIN_SILLY_BINS, BIN_LABELS[2:]), (BIN_SILLY_BINS[2:], BIN_LABELS)],
    ids=["more_bins_than_labels", "more_labels_than_bins"],
)
def test_register_binned_stratification_raises(bins, labels):
    mgr = ResultsManager()
    with pytest.raises(ValueError):
        raise mgr.register_binned_stratification(
            BIN_SOURCE, "column", BIN_BINNED_COLUMN, bins, labels
        )



import warnings
from collections import defaultdict
from typing import Callable, Dict, List, Tuple

import pandas as pd
from loguru import logger

from vivarium.framework.results.exceptions import ResultsConfigurationError
from vivarium.framework.results.stratification import Stratification


class ResultsContext:
    """
    Manager context for organizing observations and the stratifications they require.

    This context object is wholly contained by the manager :class:`vivarium.framework.results.manager.ResultsManger`.
    Stratifications can be added to the context through the manager via the
    :meth:`vivarium.framework.results.context.ResultsContext.add_observation` method.
    """

    def __init__(self):
        self._default_stratifications: List[str] = []
        self._stratifications: List[Stratification] = []
        # keys are event names: ["time_step__prepare", "time_step", "time_step__cleanup", "collect_metrics"]
        # values are dicts with key (filter, grouper) value (measure, aggregator_sources, aggregator, additional_keys)
        self._observations = defaultdict(lambda: defaultdict(list))

    # noinspection PyAttributeOutsideInit
    def set_default_stratifications(self, default_grouping_columns: List[str]):
        if self._default_stratifications:
            raise ResultsConfigurationError(
                "Multiple calls are being made to set default grouping columns "
                "for results production."
            )
        if not default_grouping_columns:
            raise ResultsConfigurationError(
                "Attempting to set an empty list as the default grouping columns "
                "for results production."
            )
        self._default_grouping_columns = default_grouping_columns

    def add_stratification(
        self,
        name: str,
        sources: List[str],
        categories: List[str],
        mapper: Callable,
        is_vectorized: bool,
    ):
        """Add a stratification to the context.

        Parameters
        ----------
        name
            Name of the of the column created by the stratification.
        sources
            A list of the columns and values needed for the mapper to determinate
            categorization.
        categories
            List of string values that the mapper is allowed to output.
        mapper
            A callable that emits values in `categories` given inputs from columns
            and values in the `requires_columns` and `requires_values`, respectively.
        is_vectorized
            `True` if the mapper function expects a `DataFrame`, and `False` if it
            expects a row of the `DataFrame` and should be used by calling :func:`df.apply`.


        Returns
        ------
        None
        """
        if len([s.name for s in self._stratifications if s.name == name]):
            raise ValueError(f"Name `{name}` is already used")
        stratification = Stratification(name, sources, categories, mapper, is_vectorized)
        self._stratifications.append(stratification)

    def add_observation(
        self,
        name: str,
        pop_filter: str,
        aggregator_sources: List[str],
        aggregator: Callable[[pd.DataFrame], float],
        additional_stratifications: List[str] = (),
        excluded_stratifications: List[str] = (),
        when: str = "collect_metrics",
        **additional_keys: str,
    ):
        self._warn_check_stratifications(additional_stratifications, excluded_stratifications)
        stratifications = self._get_stratifications(
            additional_stratifications, excluded_stratifications
        )
        self._observations[when][(pop_filter, stratifications)].append(
            (name, aggregator_sources, aggregator, additional_keys)
        )

    def gather_results(self, population: pd.DataFrame, event_name: str) -> Dict[str, float]:
        # Optimization: We store all the producers by pop_filter and stratifications
        # so that we only have to apply them once each time we compute results.
        for stratification in self._stratifications:
            population = stratification(population)

        for (pop_filter, stratifications), observations in self._observations[
            event_name
        ].items():
            # Results production can be simplified to
            # filter -> groupby -> aggregate in all situations we've seen.
            if pop_filter:
                population = population.query(pop_filter)
            pop_groups = population.groupby(list(stratifications))
            for measure, aggregator_sources, aggregator, additional_keys in observations:
                if aggregator_sources:
                    aggregates = pop_groups[aggregator_sources].apply(aggregator).fillna(0.0)
                else:
                    aggregates = pop_groups.apply(aggregator)

                # Ensure we are dealing with a single column of formattable results
                if isinstance(aggregates, pd.DataFrame):
                    aggregates = aggregates.squeeze(axis=1)
                if not isinstance(aggregates, pd.Series):
                    raise TypeError(
                        f"The aggregator return value is a {type(aggregates)} and could not be "
                        "made into a pandas.Series. This is probably not correct."
                    )

                # Keep formatting all in one place.
                yield self._format_results(measure, aggregates, **additional_keys)

    def _get_stratifications(
        self,
        additional_stratifications: List[str] = (),
        excluded_stratifications: List[str] = (),
    ) -> Tuple[str, ...]:
        stratifications = list(
            set(self._default_stratifications) - set(excluded_stratifications)
            | set(additional_stratifications)
        )
        # Makes sure measure identifiers have fields in the same relative order.
        return tuple(sorted(stratifications))

    @staticmethod
    def _format_results(
        measure: str,
        aggregates: pd.Series,
        **additional_keys: str,
    ) -> Dict[str, float]:
        results = {}
        # First we expand the categorical index over unobserved pairs.
        # This ensures that the produced results are always the same length.
        if isinstance(aggregates.index, pd.MultiIndex):
            idx = pd.MultiIndex.from_product(
                aggregates.index.levels, names=aggregates.index.names
            )
        else:
            idx = aggregates.index
        data = pd.Series(data=0, index=idx)
        data.loc[aggregates.index] = aggregates

        def _format(field, param):
            """Format of the measure identifier tokens into FIELD_param."""
            return f"{str(field).upper()}_{param}"

        for categories, val in data.items():
            if isinstance(categories, str):  # handle single stratification case
                categories = [categories]
            key = "_".join(
                [_format("measure", measure)]
                + [
                    _format(field, category)
                    for field, category in zip(data.index.names, categories)
                ]
                # Sorts additional_keys by the field name.
                + [
                    _format(field, category)
                    for field, category in sorted(additional_keys.items())
                ]
            )
            results[key] = val
        return results

    def _warn_check_stratifications(
        self, additional_stratifications, excluded_stratifications
    ):
        """Check additional and excluded stratifications if they'd not affect
        stratifications (i.e., would be NOP), and emit warning."""
        nop_additional = [
            s for s in additional_stratifications if s in self._default_stratifications
        ]
        if len(nop_additional):
            warnings.warn(
                f"Specified additional stratifications are already included by default: {nop_additional}",
                UserWarning,
            )
        nop_exclude = [
            s for s in excluded_stratifications if s not in self._default_stratifications
        ]
        if len(nop_exclude):
            warnings.warn(
                f"Specified excluded stratifications are already not included by default: {nop_exclude}",
                UserWarning,
            )
        if len(nop_additional) or len(nop_exclude):
            logger.debug(f"Default stratifications are: {self._default_stratifications}")



"""
===============
The Config Tree
===============

A configuration structure which supports cascading layers.

In ``vivarium`` it allows base configurations to be overridden by component
level configurations which are in turn overridden by model level configuration
which can be overridden by user supplied overrides. From the perspective
of normal client code the cascading is hidden and configuration values
are presented as attributes of the configuration object the values of
which are the value of that key in the outermost layer of configuration
where it appears.

For example:

.. code-block:: python

    >>> config = ConfigTree(layers=['inner_layer', 'middle_layer', 'outer_layer', 'user_overrides'])
    >>> config.update({'section_a': {'item1': 'value1', 'item2': 'value2'}, 'section_b': {'item1': 'value3'}}, layer='inner_layer')
    >>> config.update({'section_a': {'item1': 'value4'}, 'section_b': {'item1': 'value5'}}, layer='middle_layer')
    >>> config.update({'section_b': {'item1': 'value6'}}, layer='outer_layer')
    >>> config.section_a.item1
    'value4'
    >>> config.section_a.item2
    'value2'
    >>> config.section_b.item1
    'value6'

"""
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import yaml

from vivarium.exceptions import VivariumError


class ConfigurationError(VivariumError):
    """Base class for configuration errors."""

    def __init__(self, message: str, value_name: Optional[str]):
        self.value_name = value_name
        super().__init__(message)


class ConfigurationKeyError(ConfigurationError, KeyError):
    """Error raised when a configuration lookup fails."""

    pass


class DuplicatedConfigurationError(ConfigurationError):
    """Error raised when a configuration value is set more than once.

    Attributes
    ----------
    layer
        The configuration layer at which the value is being set.
    source
        The original source of the configuration value.
    value
        The original configuration value.

    """

    def __init__(
        self, message: str, name: str, layer: Optional[str], source: Optional[str], value: Any
    ):
        self.layer = layer
        self.source = source
        self.value = value
        super().__init__(message, name)


class ConfigNode:
    """A priority based configuration value.

    A :class:`ConfigNode` represents a single configuration value with
    priority-based layers.  The intent is to allow a value to be set from
    sources with different priorities and to record what the value was set
    to and from where.

    For example, a simulation may need certain values to always exist, and so
    it will set them up at a "base" layer. Components in the simulation may
    have a different set of priorities and so override the "base" value at
    a "component" level.  Finally a user may want to override the simulation
    and component defaults with values at the command line or interactively,
    and so those values will be set in a final "user" layer.

    A :class:`ConfigNode` may only have a value set at each layer once.
    Attempts to set a value at the same layer multiple times will result in
    a :class:`DuplicatedConfigurationError`.

    The :class:`ConfigNode` will record all values set and the source they
    are set from.  This sort of provenance with configuration data greatly
    eases debugging and analysis of simulation code.

    This class should not be instantiated directly. All interaction should
    take place by manipulating a :class:`ConfigTree` object.

    """

    def __init__(self, layers: List[str], name: str):
        self._name = name
        self._layers = layers
        self._values = {}
        self._frozen = False
        self._accessed = False

    @property
    def name(self) -> str:
        """The name of this configuration value."""
        return self._name

    @property
    def accessed(self) -> bool:
        """Returns whether this node has been accessed."""
        return self._accessed

    @property
    def metadata(self) -> List[Dict[str, Any]]:
        """Returns all values and associated metadata for this node."""
        result = []
        for layer in self._layers:
            if layer in self._values:
                result.append(
                    {
                        "layer": layer,
                        "source": self._values[layer][0],
                        "value": self._values[layer][1],
                    }
                )
        return result

    def freeze(self):
        """Causes the :class:`ConfigNode` node to become read only.

        This can be used to create a contract around when the configuration is
        modifiable.

        """
        self._frozen = True

    def get_value(self, layer: Optional[str]) -> Any:
        """Returns the value at the specified layer.

        If no layer is specified, the outermost (highest priority) layer
        at which a value has been set will be used.

        Parameters
        ----------
        layer
            Name of the layer to retrieve the value from.

        Raises
        ------
        KeyError
            If no value has been set at any layer.

        """
        value = self._get_value_with_source(layer)[1]
        self._accessed = True
        return value

    def update(self, value: Any, layer: Optional[str], source: Optional[str]):
        """Set a value for a layer with optional metadata about source.

        Parameters
        ----------
        value
            Data to store in the node.
        layer
            Name of the layer to use. If no layer is provided, the value will
            be set in the outermost (highest priority) layer.
        source
            Metadata indicating the source of this value.

        Raises
        ------
        ConfigurationError
            If the node is frozen.
        ConfigurationKeyError
            If the provided layer does not exist.
        DuplicatedConfigurationError
            If a value has already been set at the provided layer or a value
            is already in the outermost layer and no layer has been provided.

        """
        if self._frozen:
            raise ConfigurationError(
                f"Frozen ConfigNode {self.name} does not support assignment.", self.name
            )

        layer = layer if layer else self._layers[-1]

        if layer not in self._layers:
            raise ConfigurationKeyError(
                f"No layer {layer} in ConfigNode {self.name}.", self.name
            )
        elif layer in self._values:
            source, value = self._values[layer]
            raise DuplicatedConfigurationError(
                f"Value has already been set at layer {layer}.",
                name=self.name,
                layer=layer,
                source=source,
                value=value,
            )
        else:
            self._values[layer] = (source, value)

    def _get_value_with_source(self, layer: Optional[str]) -> Tuple[str, Any]:
        """Returns a (source, value) tuple at the specified layer.

        If no layer is specified, the outermost (highest priority) layer
        at which a value has been set will be used.

        Parameters
        ----------
        layer
            Name of the layer to retrieve the (source, value) pair from.

        Raises
        ------
        KeyError
            If no value has been set at any layer.

        """
        if layer and layer in self._values:
            return self._values[layer]

        for layer in reversed(self._layers):
            if layer in self._values:
                return self._values[layer]

        raise ConfigurationKeyError(
            f"No value stored in this ConfigNode {self.name}.", self.name
        )

    def __bool__(self):
        return bool(self._values)

    def __repr__(self):
        out = []
        for m in reversed(self.metadata):
            layer, source, value = m.values()
            out.append(f"{layer}: {value}\n    source: {source}")
        return "\n".join(out)

    def __str__(self):
        if not self:
            return ""
        layer, _, value = self.metadata[-1].values()
        return f"{layer}: {value}"


class ConfigTree:
    """A container for configuration information.

    Each configuration value is exposed as an attribute the value of which
    is determined by the outermost layer which has the key defined.

    """

    def __init__(
        self,
        data: Union[Dict, str, Path, "ConfigTree"] = None,
        layers: List[str] = None,
        name: str = "",
    ):
        """
        Parameters
        ----------
        data
            The :class:`ConfigTree` accepts many kinds of data:

             - :class:`dict` : Flat or nested dictionaries may be provided.
               Keys of dictionaries at all levels must be strings.
             - :class:`ConfigTree` : Another :class:`ConfigTree` can be
               used. All source information will be ignored and the source
               will be set to 'initial_data' and values will be stored at
               the lowest priority level.
             - :class:`str` : Strings provided can be yaml formatted strings,
               which will be parsed into a dictionary using standard yaml
               parsing. Alternatively, a path to a yaml file may be provided
               and the file will be read in and parsed.
             - :class:`pathlib.Path` : A path object to a yaml file will
               be interpreted the same as a string representation.

            All values will be set with 'initial_data' as the source and
            will use the lowest priority level. If values are set at higher
            priorities they will be used when the :class:`ConfigTree` is
            accessed.
        layers
            A list of layer names. The order in which layers defined
            determines their priority.  Later layers override the values from
            earlier ones.

        """
        self.__dict__["_layers"] = layers if layers else ["base"]
        self.__dict__["_children"] = {}
        self.__dict__["_frozen"] = False
        self.__dict__["_name"] = name
        self.update(data, layer=self._layers[0], source="initial data")

    def freeze(self):
        """Causes the ConfigTree to become read only.

        This is useful for loading and then freezing configurations that
        should not be modified at runtime.

        """
        self.__dict__["_frozen"] = True
        for child in self.values():
            child.freeze()

    def items(self) -> Iterable[Tuple[str, Union["ConfigTree", ConfigNode]]]:
        """Return an iterable of all (child_name, child) pairs."""
        return self._children.items()

    def keys(self) -> Iterable[str]:
        """Return an Iterable of all child names."""
        return self._children.keys()

    def values(self) -> Iterable:
        """Return an Iterable of all children."""
        return self._children.values()

    def unused_keys(self) -> List[str]:
        """Lists all values in the ConfigTree that haven't been accessed."""
        unused = []
        for name, child in self.items():
            if isinstance(child, ConfigNode):
                if not child.accessed:
                    unused.append(name)
            else:
                for grandchild_name in child.unused_keys():
                    unused.append(f"{name}.{grandchild_name}")
        return unused

    def to_dict(self) -> Dict:
        """Converts the ConfigTree into a nested dictionary.

        All metadata is lost in this conversion.

        """
        result = {}
        for name, child in self.items():
            if isinstance(child, ConfigNode):
                result[name] = child.get_value(layer=None)
            else:
                result[name] = child.to_dict()
        return result

    def get_from_layer(self, name: str, layer: str = None) -> Any:
        """Get a configuration value from the provided layer.

        If no layer is specified, the outermost (highest priority) layer
        at which a value has been set will be used.

        Parameters
        ----------
        name
            The name of the value to retrieve
        layer
            The name of the layer to retrieve the value from.

        """
        if name not in self:
            name = f"{self._name}.{name}" if self._name else name
            raise ConfigurationKeyError(f"No value at name {name}.", name)

        child = self._children[name]
        if isinstance(child, ConfigNode):
            return child.get_value(layer)
        else:
            return child

    def update(
        self,
        data: Union[Dict, str, Path, "ConfigTree", None],
        layer: str = None,
        source: str = None,
    ):
        """Adds additional data into the :class:`ConfigTree`.

        Parameters
        ----------
        data
            :func:`~ConfigTree.update` accepts many types of data.

             - :class:`dict` : Flat or nested dictionaries may be provided.
               Keys of dictionaries at all levels must be strings.
             - :class:`ConfigTree` : Another :class:`ConfigTree` can be
               used. All source information will be ignored and the
               provided layer and source will be used to set the metadata.
             - :class:`str` : Strings provided can be yaml formatted strings,
               which will be parsed into a dictionary using standard yaml
               parsing. Alternatively, a path to a yaml file may be provided
               and the file will be read in and parsed.
             - :class:`pathlib.Path` : A path object to a yaml file will
               be interpreted the same as a string representation.
        layer
            The name of the layer to store the value in.  If no layer is
            provided, the value will be set in the outermost (highest priority)
            layer.
        source
            The source to attribute the value to.

        Raises
        ------
        ConfigurationError
            If the :class:`ConfigTree` is frozen or attempting to assign
            an invalid value.
        ConfigurationKeyError
            If the provided layer does not exist.
        DuplicatedConfigurationError
            If a value has already been set at the provided layer or a value
            is already in the outermost layer and no layer has been provided.

        """
        if data is not None:
            data, source = self._coerce(data, source)
            for k, v in data.items():
                self._set_with_metadata(k, v, layer, source)

    def metadata(self, name: str) -> List[Dict[str, Any]]:
        if name in self:
            return self._children[name].metadata
        name = f"{self._name}.{name}" if self._name else name
        raise ConfigurationKeyError(f"No configuration value with name {name}", name)

    @staticmethod
    def _coerce(
        data: Union[Dict, str, Path, "ConfigTree"], source: Union[str, None]
    ) -> Tuple[Dict, Union[str, None]]:
        """Coerces data into dictionary format."""
        if isinstance(data, dict):
            return data, source
        elif (isinstance(data, str) and data.endswith((".yaml", ".yml"))) or isinstance(
            data, Path
        ):
            source = source if source else str(data)
            with open(data) as f:
                data = f.read()
            data = yaml.full_load(data)
            return data, source
        elif isinstance(data, str):
            data = yaml.full_load(data)
            return data, source
        elif isinstance(data, ConfigTree):
            return data.to_dict(), source
        else:
            raise ConfigurationError(
                f"ConfigTree can only update from dictionaries, strings, paths and ConfigTrees. "
                f"You passed in {type(data)}",
                value_name=None,
            )

    def _set_with_metadata(
        self, name: str, value: Any, layer: Optional[str], source: Optional[str]
    ):
        """Set a value in the named layer with the given source.

        Parameters
        ----------
        name
            The name of the value.
        value
            The value to store.
        layer
            The name of the layer to store the value in.  If no layer is
            provided, the value will be set in the outermost (highest priority)
            layer.
        source
            The source to attribute the value to.

        Raises
        ------
        ConfigurationError
            If the :class:`ConfigTree` is frozen or attempting to assign
            an invalid value.
        ConfigurationKeyError
            If the provided layer does not exist.
        DuplicatedConfigurationError
            If a value has already been set at the provided layer or a value
            is already in the outermost layer and no layer has been provided.

        """
        if self._frozen:
            raise ConfigurationError(
                f"Frozen ConfigTree {self._name} does not support assignment.", self._name
            )

        if isinstance(value, dict):
            if name not in self:
                self._children[name] = ConfigTree(layers=list(self._layers), name=name)
            if isinstance(self._children[name], ConfigNode):
                name = f"{self._name}.{name}" if self._name else name
                raise ConfigurationError(
                    f"Can't assign a dictionary as a value to a ConfigNode.", name
                )
        else:
            if name not in self:
                self._children[name] = ConfigNode(list(self._layers), name=self._name)
            if isinstance(self._children[name], ConfigTree):
                name = f"{self._name}.{name}" if self._name else name
                raise ConfigurationError(f"Can't assign a value to a ConfigTree.", name)

        self._children[name].update(value, layer, source)

    def __setattr__(self, name, value):
        """Set a value on the outermost layer."""
        if name not in self:
            raise ConfigurationKeyError(
                "New configuration keys can only be created with the update method.",
                self._name,
            )
        self._set_with_metadata(name, value, layer=None, source=None)

    def __setitem__(self, name, value):
        """Set a value on the outermost layer."""
        if name not in self:
            raise ConfigurationKeyError(
                "New configuration keys can only be created with the update method.",
                self._name,
            )
        self._set_with_metadata(name, value, layer=None, source=None)

    def __getattr__(self, name):
        """Get a value from the outermost layer in which it appears."""
        return self.get_from_layer(name)

    def __getitem__(self, name):
        """Get a value from the outermost layer in which it appears."""
        return self.get_from_layer(name)

    def __delattr__(self, name):
        if name in self:
            del self._children[name]

    def __delitem__(self, name):
        if name in self:
            del self._children[name]

    def __contains__(self, name):
        """Test if a configuration key exists in any layer."""
        return name in self._children

    def __iter__(self):
        """Dictionary-like iteration."""
        return iter(self._children)

    def __len__(self):
        return len(self._children)

    def __dir__(self):
        return list(self._children.keys()) + dir(super(ConfigTree, self))

    def __repr__(self):
        return "\n".join(
            [
                "{}:\n    {}".format(name, repr(c).replace("\n", "\n    "))
                for name, c in self._children.items()
            ]
        )

    def __str__(self):
        return "\n".join(
            [
                "{}:\n    {}".format(name, str(c).replace("\n", "\n    "))
                for name, c in self._children.items()
            ]
        )



import itertools
import math

import numpy as np
import pandas as pd

##########################
# Mock data and fixtures #
##########################
NAME = "hogwarts_house"
SOURCES = ["first_name", "last_name"]
CATEGORIES = ["hufflepuff", "ravenclaw", "slytherin", "gryffindor"]
STUDENT_TABLE = pd.DataFrame(
    np.array([["harry", "potter"], ["severus", "snape"], ["luna", "lovegood"]]),
    columns=SOURCES,
)
STUDENT_HOUSES = pd.Series(["gryffindor", "slytherin", "ravenclaw"])

BIN_BINNED_COLUMN = "silly_bin"
BIN_SOURCE = "silly_level"
BIN_LABELS = ["not", "meh", "somewhat", "very", "extra"]
BIN_SILLY_BINS = [0, 20, 40, 60, 90]

COL_NAMES = ["house", "familiar", "power_level", "tracked"]
FAMILIARS = ["owl", "cat", "gecko", "banana_slug", "unladen_swallow"]
POWER_LEVELS = [i * 10 for i in range(5, 9)]
TRACKED_STATUSES = [True, False]
RECORDS = [
    (house, familiar, power_level, ts)
    for house, familiar, power_level, ts in itertools.product(
        CATEGORIES, FAMILIARS, POWER_LEVELS, TRACKED_STATUSES
    )
]
BASE_POPULATION = pd.DataFrame(data=RECORDS, columns=COL_NAMES)


##################
# Helper methods #
##################
def sorting_hat_vector(state_table: pd.DataFrame) -> pd.Series:
    sorted_series = state_table.apply(sorting_hat_serial, axis=1)
    return sorted_series


def sorting_hat_serial(simulant_row: pd.Series) -> str:
    first_name = simulant_row[0]
    last_name = simulant_row[1]
    if first_name == "harry":
        return "gryffindor"
    if first_name == "luna":
        return "ravenclaw"
    if last_name == "snape":
        return "slytherin"
    return "hufflepuff"


def sorting_hat_bad_mapping(simulant_row: pd.Series) -> str:
    # Return something not in CATEGORIES
    return "pancakes"


def verify_stratification_added(
    stratification_list, name, sources, categories, mapper, is_vectorized
):
    """Verify that a :class: `vivarium.framework.results.stratification.Stratification` is in `stratification_list`"""
    matching_stratification_found = False
    for stratification in stratification_list:  # noqa
        # big equality check
        if (
            stratification.name == name
            and sorted(stratification.categories) == sorted(categories)
            and stratification.mapper == mapper
            and stratification.is_vectorized == is_vectorized
            and sorted(stratification.sources) == sorted(sources)
        ):
            matching_stratification_found = True
            break
    return matching_stratification_found



import pytest
import yaml

from vivarium.framework.components.parser import (
    ComponentConfigurationParser,
    ParsingError,
    import_and_instantiate_components,
    parse_component_config_to_list,
    prep_components,
)
from vivarium.framework.configuration import build_simulation_configuration

from .mocks import MockComponentA, MockComponentB

TEST_COMPONENTS_NESTED = """
components:
    ministry.silly_walk:
       - Prance()
       - Jump('front_flip')
       - PratFall('15')
    pet_shop:
       - Parrot()
       - dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
"""

# common case when users comment out and ends up having none
TEST_COMPONENTS_BAD = """
components:
    ministry.silly_walk:
       - Prance()
       - Jump('front_flip')
       - PratFall('15')
    pet_shop:
#       - Parrot()
#       - dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
"""

TEST_COMPONENTS_FLAT = """
components:
    - ministry.silly_walk.Prance()
    - ministry.silly_walk.Jump('front_flip')
    - ministry.silly_walk.PratFall('15')
    - pet_shop.Parrot()
    - pet_shop.dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
"""

TEST_COMPONENTS_PARSED = [
    "ministry.silly_walk.Prance()",
    "ministry.silly_walk.Jump('front_flip')",
    "ministry.silly_walk.PratFall('15')",
    "pet_shop.Parrot()",
    "pet_shop.dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')",
]

TEST_COMPONENTS_PREPPED = [
    ("ministry.silly_walk.Prance", tuple()),
    ("ministry.silly_walk.Jump", ("front_flip",)),
    ("ministry.silly_walk.PratFall", ("15",)),
    ("pet_shop.Parrot", tuple()),
    ("pet_shop.dangerous_animals.Crocodile", ("gold_tooth", "teacup", "3.14")),
]


def mock_importer(path):
    return {
        "test_components.MockComponentA": MockComponentA,
        "test_components.MockComponentB": MockComponentB,
    }[path]


@pytest.fixture(params=[TEST_COMPONENTS_NESTED, TEST_COMPONENTS_FLAT])
def components(request):
    return request.param


@pytest.fixture
def import_and_instantiate_mock(mocker):
    return mocker.patch(
        "vivarium.framework.components.parser.import_and_instantiate_components"
    )


def test_parse_component_config(components):

    source = yaml.full_load(components)["components"]
    component_list = parse_component_config_to_list(source)
    assert set(TEST_COMPONENTS_PARSED) == set(component_list)


def test_prep_components():
    desc = 'cave_system.monsters.Rabbit("timid", "squeak")'
    component, args = prep_components([desc])[0]
    assert component == "cave_system.monsters.Rabbit"
    assert set(args) == {"timid", "squeak"}


def test_parse_component_syntax_error():
    desc = 'cave_system.monsters.Rabbit("timid", 0.01)'
    with pytest.raises(ParsingError):
        prep_components([desc])

    desc = 'cave_system.monsters.Rabbit("timid\', "0.01")'
    with pytest.raises(ParsingError):
        prep_components([desc])

    desc = "cave_system.monsters.Rabbit(\"timid', '0.01')"
    with pytest.raises(ParsingError):
        prep_components([desc])


def test_parse_and_prep_components(components):
    source = yaml.full_load(components)["components"]
    prepped_components = prep_components(parse_component_config_to_list(source))

    assert set(TEST_COMPONENTS_PREPPED) == set(prepped_components)


def test_import_and_instantiate_components(monkeypatch):
    monkeypatch.setattr("vivarium.framework.components.parser.import_by_path", mock_importer)

    component_descriptions = [
        ("test_components.MockComponentA", ("A Hundred and One Ways to Start a Fight",)),
        ("test_components.MockComponentB", ("Ethel the Aardvark goes Quantity Surveying",)),
    ]
    component_list = import_and_instantiate_components(component_descriptions)

    assert len(component_list) == 2
    assert isinstance(component_list[0], MockComponentA)
    assert component_list[0].args == ("A Hundred and One Ways to Start a Fight",)
    assert isinstance(component_list[1], MockComponentB)
    assert component_list[1].args == ("Ethel the Aardvark goes Quantity Surveying",)


def test_ComponentConfigurationParser_get_components(import_and_instantiate_mock, components):
    config = build_simulation_configuration()
    config.update(components)

    parser = ComponentConfigurationParser()
    parser.get_components(config.components)

    import_and_instantiate_mock.assert_called_once_with(TEST_COMPONENTS_PREPPED)


def test_components_config_valid():
    bad_config = yaml.full_load(TEST_COMPONENTS_BAD)["components"]
    with pytest.raises(ParsingError):
        parse_component_config_to_list(bad_config)
